//  Copyright (c) 2006, Gustavo Franco
//  Email:  gustavo_franco@hotmail.com
//  All rights reserved.

//  Redistribution and use in source and binary forms, with or without modification, 
//  are permitted provided that the following conditions are met:

//  Redistributions of source code must retain the above copyright notice, 
//  this list of conditions and the following disclaimer. 
//  Redistributions in binary form must reproduce the above copyright notice, 
//  this list of conditions and the following disclaimer in the documentation 
//  and/or other materials provided with the distribution. 

//  THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY
//  KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
//  IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A PARTICULAR
//  PURPOSE. IT CAN BE DISTRIBUTED FREE OF CHARGE AS LONG AS THIS HEADER 
//  REMAINS UNCHANGED.

using System;
using System.IO;
using System.Data;
using System.Text;
using System.Drawing;
using System.Windows.Forms;
using System.ComponentModel;
using System.Drawing.Imaging;
using System.Collections.Generic;
using System.Runtime.InteropServices;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;


using CustomControls.Controls;

namespace CustomControls
{   
    public partial class OpenTextureDialog : OpenFileDialogEx
    {
        static string ImageSupportedExtensions = ".bmp.tga.dds.png.jpeg.jpg.gif.tif.tiff.hdr.exr";

        IO.ImagePreviewRenderer renderer = null;
        bool bUseNormalMapping = false;
        string textureFile;
        string title;

        #region Constructors
        public OpenTextureDialog(bool _bUseNormalMapping,string _textureFile)
        {
            InitializeComponent();
            textureFile = _textureFile;
            bUseNormalMapping = _bUseNormalMapping;
            title = this.Text;
        }
        #endregion

        #region Overrides
        /// <summary>
        /// t@CύXꂽƂ
        /// </summary>
        /// <param name="filePath">t@C</param>
        public override void OnFileNameChanged(string filePath)
        {
            FileAttributes  attr = File.GetAttributes(filePath);            
            if((attr & FileAttributes.Directory) == FileAttributes.Directory) return;

            if(ImageSupportedExtensions.IndexOf(ImageSupportedExtensions)>=0)
            {
                if(bUseNormalMapping)
                    renderer.NormalMapFile = filePath;
                else
                    renderer.TextureImageFile = filePath;
            }
        }

        /// <summary>
        /// tH_ύXꂽƂ
        /// </summary>
        /// <param name="folderName">tH_</param>
        public override void OnFolderNameChanged(string folderName)
        {
        }

        protected override void OnLoad(EventArgs e)
        {
            base.OnLoad(e);

            try
            {
                renderer = new IO.ImagePreviewRenderer();
                renderer.InitializeGraphics(pbxPreview);
                if(bUseNormalMapping)
                    renderer.TextureImageFile = textureFile;
            }
            catch (Exception ex)
            {
                MessageBox.Show("G[܂BfBXvC̐ݒĂB" + ex.ToString());
                return;
            }

            timer1.Start();
        }

        public override void OnClosingDialog()
        {
            timer1.Stop();
            if(renderer!=null)
            {
                renderer.TerminateGraphics();
                renderer = null;
            }

            base.OnClosingDialog();
        }
        #endregion

        private void pbxPreview_Paint(object sender, PaintEventArgs e)
        {
            if(renderer!=null)
                renderer.Render(bUseNormalMapping);
        }

        private void timer1_Tick(object sender, EventArgs e)
        {
            if(renderer!=null && renderer.isAvailable())
                renderer.Render(bUseNormalMapping);

        }
    }
}