/////////////////////////////////////////////////////////
//
// Image Preview Renderer
//
// Copyright 2007 Shintaro Takemura
//

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Drawing.Imaging;
using System.Text;
using System.IO;
using System.Windows.Forms;
using System.Runtime.InteropServices;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

namespace IO
{
    public class ImagePreviewRenderer
    {
        /// <summary>
        /// Direct3D foCX
        /// </summary>
        private Device device = null;
        private PresentParameters presentParameters = null;

        // QctHg
        private Microsoft.DirectX.Direct3D.Font font = null;

        // J[eNX`
        private Texture g_baseTexture = null;

        // @{nCg}bveNX`
        private Texture g_nmhTexture = null;

        // |XgGtFNgpeNX`igpj
        private Texture ds4Texture, ds16Texture, hBlurTexture, vBlurTexture;

        // r{[h̒_obt@i̋`j
        private CustomVertex.TransformedColored[] backgroundVertices = new CustomVertex.TransformedColored[4];

        // r{[h̃CfbNXobt@i̋`j
        private short[] backgroundIndices;

        // bV
        private Mesh boxMesh = null;

        // GtFNg
        private Effect effect = null;

        // AXyNg
        private Size aspect = new Size(1,1);

        /// <summary>
        /// eNX`摜
        /// </summary>
        private string textureImageFile = null;

        public string TextureImageFile
        {
            get { return textureImageFile; }
            set {
                textureImageFile = value;
                if(device==null) return;
                try
                {
                    if(g_baseTexture!=null) g_baseTexture.Dispose();
                    if(!string.IsNullOrEmpty(textureImageFile))
                    {
                        ImageInformation	info = TextureLoader.ImageInformationFromFile(textureImageFile);
                        StringBuilder sb = new StringBuilder();
                        sb.Append(info.Width);
                        sb.Append(" x ");
                        sb.Append(info.Height);
                        sb.Append(" , ");
                        sb.Append(info.Format.ToString());                        
                        Usage = sb.ToString();
                        g_baseTexture = TextureLoader.FromFile(device,textureImageFile);
                        aspect = new Size(info.Width,info.Height);
                    }
                }
                catch{}
            }
        }

        /// <summary>
        /// @}bv
        /// </summary>
        private string normalMapFile = null;

        public string NormalMapFile
        {
            get { return normalMapFile; }
            set { normalMapFile = value;
                if(device==null) return;
                try
                {
                    if(g_nmhTexture!=null) g_nmhTexture.Dispose();
                    if(!string.IsNullOrEmpty(normalMapFile))
                        g_nmhTexture = TextureLoader.FromFile(device,normalMapFile);           
                }
                catch {}
            }
        }

        /// <summary>
        /// bZ[W
        /// </summary>
        private string usage;

        public string Usage
        {
            get { return usage; }
            set { usage = value; }
        }

        /// <summary>
        /// p\
        /// </summary>
        /// <returns></returns>
        public bool isAvailable()
        {
            return device!=null && !device.Disposed;
        }

        /// <summary>
        /// OtBbNX֘Ã\[X
        /// </summary>
        public void InitializeGraphics(Control control)
        {
            presentParameters = new PresentParameters();

            presentParameters.Windowed = true;
            presentParameters.BackBufferWidth = control.ClientSize.Width;
            presentParameters.BackBufferHeight = control.ClientSize.Height;
            presentParameters.BackBufferFormat = Manager.Adapters.Default.CurrentDisplayMode.Format;
            presentParameters.BackBufferCount = 1;
            presentParameters.DeviceWindowHandle = control.Handle;
            presentParameters.SwapEffect = SwapEffect.Discard;

            presentParameters.EnableAutoDepthStencil = true;
            presentParameters.AutoDepthStencilFormat = DepthFormat.D16;

            // ł邾MSAAg
            for(MultiSampleType msType = MultiSampleType.EightSamples; msType>=MultiSampleType.None; msType--)
		    {
			    int result, quality;
			    if (Manager.CheckDeviceMultiSampleType(
                    Manager.Adapters.Default.Adapter,
				    DeviceType.Hardware,
                    Manager.Adapters[Manager.Adapters.Default.Adapter].CurrentDisplayMode.Format,
				    true,
                    msType,
                    out result, out quality))
			    {
				    presentParameters.MultiSample = msType;
				    presentParameters.MultiSampleQuality = 0;
                    break;
			    }
		    }

            presentParameters.FullScreenRefreshRateInHz = 0;
            presentParameters.PresentationInterval = PresentInterval.Default;

            // Direct3D foCX̐
            this.device = new Device(
                Manager.Adapters.Default.Adapter,
                DeviceType.Hardware,
                control.Handle,
                CreateFlags.HardwareVertexProcessing, presentParameters);

            // foCXZbg̃Cxgnh̐ݒ
            this.device.DeviceReset += this.OnResetDevice;

            // foCXXg̃Cxgnh̐ݒ
            this.device.DeviceLost += this.OnLostDevice;

            // foCXTCỸCxgnh̐ݒ
            this.device.DeviceResizing += OnDeviceResizing;

            string error;
            // GtFNgt@Cǂݍ
            ShaderFlags flag = ShaderFlags.None;
            effect = Effect.FromFile(
                this.device, Path.Combine(
                Path.GetDirectoryName(Application.ExecutablePath), @"TextureDialog.fxo"),
                null, null, flag, null, out error);
            if (effect == null)
            {
                MessageBox.Show(error);
                return;
            }

            // \[Xmۂ̂߂ɈӐ}IɃfoCXZbgCxgs
            this.OnResetDevice(this,null);

        }

        /// <summary>
        /// OtBbNX֘A\[X
        /// </summary>
        public void TerminateGraphics()
        {
            if (this.device != null)
            {
                // \[XĴ߂ɈӐ}IɃfoCXZbgCxgs
                OnLostDevice(this, null);

                // foCXZbg̃Cxgnh̉
                this.device.DeviceReset -= this.OnResetDevice;

                // foCXXg̃Cxgnh̉
                this.device.DeviceLost -= this.OnLostDevice;

                // foCXTCỸCxgnh̉
                this.device.DeviceResizing -= OnDeviceResizing;

                // Direct3D foCX
                this.device.Dispose();
                this.device = null;

                // GtFNg
                effect.Dispose();
                effect = null;               
            }
        }

        /// <summary>
        /// foCXTCỸCxgnh
        /// </summary>
        private void OnDeviceResizing(object sender, System.ComponentModel.CancelEventArgs e)
        {
            e.Cancel = true;
            return;
        }

        /// <summary>
        ///  foCXXg̃Cxgnh
        /// </summary>
        private void OnLostDevice(object sender, EventArgs e)
        {
			if(device!=null && effect!=null)
			{
                effect.OnLostDevice();
			}

            if(g_baseTexture!=null)
            {
                g_baseTexture.Dispose();
                g_baseTexture = null;
            }

            if(g_nmhTexture!=null)
            {
                g_nmhTexture.Dispose();
                g_nmhTexture = null;
            }

            if (ds4Texture != null)
                ds4Texture.Dispose();

            if (ds16Texture != null)
                ds16Texture.Dispose();

            if (hBlurTexture != null)
                hBlurTexture.Dispose();

            if (vBlurTexture != null)
                vBlurTexture.Dispose();

            if (boxMesh != null)
            {
                boxMesh.Dispose();
                boxMesh = null;
            }

            if (font != null)
            {
                font.Dispose();
                font = null;
            }
        }

        /// <summary>
        ///  foCXZbg̃Cxgnh
        /// </summary>
        private void OnResetDevice(object sender, EventArgs e)
        {
            ///// tHg̍쐬 /////
            FontDescription fontDescription = new FontDescription();
            fontDescription.Height = 12;
            fontDescription.FaceName = "Arial";
            fontDescription.Quality = FontQuality.ClearType;

            // tHg쐬
            font = new Microsoft.DirectX.Direct3D.Font(this.device, fontDescription);

            ///// bV̍쐬 /////
            // ̂̍쐬
            boxMesh = createTransformableBoxMesh();

            ///// J[eNX`̍쐬 /////
            if(!string.IsNullOrEmpty(textureImageFile))
            {
                g_baseTexture = TextureLoader.FromFile(device,textureImageFile);
            }

            ///// @{nCg}bv̍쐬 /////
            if(!string.IsNullOrEmpty(normalMapFile))
            {
                g_nmhTexture = TextureLoader.FromFile(device,normalMapFile);
            }

            /////// 1/4 kTvOeNX`̍쐬 /////
            //ds4Texture = new Texture(device,
            //    ClientSize.Width / 4, ClientSize.Height / 4,
            //    1, Usage.RenderTarget, Format.A16B16G16R16F , Pool.Default);

            /////// 1/16 kTvOeNX`̍쐬 /////
            //ds16Texture = new Texture(device,
            //    ClientSize.Width / 16, ClientSize.Height / 16,
            //    1, Usage.RenderTarget, Format.A16B16G16R16F, Pool.Default);

            /////// u[eNX`̍쐬 /////
            //hBlurTexture = new Texture(device,
            //    ClientSize.Width / 16, ClientSize.Height / 16,
            //    1, Usage.RenderTarget, Format.A16B16G16R16F, Pool.Default);

            /////// u[eNX`̍쐬 /////
            //vBlurTexture = new Texture(device,
            //    ClientSize.Width / 16, ClientSize.Height / 16,
            //    1, Usage.RenderTarget, Format.A16B16G16R16F, Pool.Default);


            float width = device.PresentationParameters.BackBufferWidth;
            float height = device.PresentationParameters.BackBufferHeight;

            // r{[h̒_ECfbNXobt@쐬
            backgroundVertices[0].Position = new Vector4(0, 0, 1f, 1f);
            backgroundVertices[0].Color = Color.FromArgb(255, 64, 64, 64).ToArgb();

            backgroundVertices[1].Position = new Vector4(width, 0, 1f, 1f);
            backgroundVertices[1].Color = Color.FromArgb(255, 64, 64, 64).ToArgb();

            backgroundVertices[2].Position = new Vector4(0, height, 1f, 1f);
            backgroundVertices[2].Color = Color.FromArgb(255, 0, 0, 0).ToArgb();

            backgroundVertices[3].Position = new Vector4(width, height, 1f, 1f);
            backgroundVertices[3].Color = Color.FromArgb(255, 0, 0, 0).ToArgb();

            backgroundIndices = new short[] { 0, 1, 2, 1, 3, 2 };

            if(effect!=null)
            {
                effect.OnResetDevice();
            }
        }

        /// <summary>
        /// ڋԊxNgQ_\
        /// </summary>
        struct VertexTNB
        {
            public Vector3 Position;	// ʒu
            public Vector3 Tangent; 	// ڐ
            public Vector3 Normal;		// @
            public Vector3 Binormal;	// ]@
        }

        /// <summary>
        /// ό`\ȗ̂̃bV쐬
        /// </summary>
        private Mesh createTransformableBoxMesh()
        {
            #region CfbNXf[^̔z쐬
            Int16[] indexData ={
                        2,1,0,1,2,3,
                        5,6,4,6,5,7,
                        0,1,4,5,4,1,
                        6,3,2,7,3,6,
                        4,2,0,2,4,6,
                        1,3,5,7,5,3,
                      };
            #endregion

            #region x[X_f[^
            Vector3[] bpos = 
            {
                new Vector3(1, 1, 1),
                new Vector3(1, 1, -1),
                new Vector3(1, -1, 1),
                new Vector3(1, -1, -1),
                new Vector3(-1, 1, 1),
                new Vector3(-1, 1, -1),
                new Vector3(-1, -1, 1),
                new Vector3(-1, -1, -1)
            };
            #endregion

            #region @xNg
            Vector3[] bNormals =
            {
                new Vector3( 0, 0,-1),
			    new Vector3(-1, 0, 0),
			    new Vector3( 1, 0, 0),
			    new Vector3( 0, 1, 0),
			    new Vector3( 0,-1, 0),
			    new Vector3( 0, 0, 1)
            };
            #endregion

            #region ڐxNg
            Vector3[] bTangents =
            {
		        new Vector3( 1, 0, 0),
			    new Vector3( 0, 0,-1),
			    new Vector3( 0, 0, 1),
			    new Vector3( 1, 0, 0),
			    new Vector3( 1, 0, 0),
			    new Vector3(-1, 0, 0)
            };
            #endregion

            #region ]@xNg
            Vector3[] bBinormals =
            {
                new Vector3( 0, 1, 0),
			    new Vector3( 0, 1, 0),
			    new Vector3( 0, 1, 0),
			    new Vector3( 0, 0,-1),
			    new Vector3( 0, 0, 1),
			    new Vector3( 0, 1, 0)
            };
            #endregion

            #region CfbNX璸_WJ
            //_f[^̔z쐬
            VertexTNB[] verts = new VertexTNB[36];

            for (int t = 0; t < verts.Length/3; t++)
            {
                Vector3[] apex = new Vector3[3];

                // Op`PʂŒ_Zbg
                for (int a = 0; a < 3; a++)
                {
                    int n = t * 3 + a;
                    apex[a] = bpos[indexData[n]];
                    verts[n].Position = apex[a];
                }

                // @xNg̎Zo
                Vector3 normal = Vector3.Normalize(
                    Vector3.Cross(
                    Vector3.Normalize(apex[1] - apex[0]),
                    Vector3.Normalize(apex[2] - apex[0]))
                    );

                // ʎZo
                int basis = 0;
                float max_a = -1.0f;
                for (int n = 0; n < 6; n++)
                {
                    // @xNgɊʂZo
                    float a = Vector3.Dot(normal, bNormals[n]);
                    if (max_a < a)
                    {
                        max_a = a;
                        basis = n;
                    }
                }

                // @vZ
                for (int a = 0; a < 3; a++)
                {
                    int n = t * 3 + a;
                    verts[n].Normal = normal;
                    verts[n].Tangent = bTangents[basis];
                    verts[n].Binormal = bBinormals[basis];
                }
            }
            #endregion

            #region _vf̍쐬
            DeclarationType f3 = DeclarationType.Float3;
            DeclarationMethod dd = DeclarationMethod.Default;
            VertexElement[] elements =
		    {
			    new VertexElement(0, 0,  f3, dd, DeclarationUsage.Position, 0),
			    new VertexElement(0, 12, f3, dd, DeclarationUsage.Tangent, 0),
			    new VertexElement(0, 24, f3, dd, DeclarationUsage.Normal, 0),
			    new VertexElement(0, 36, f3, dd, DeclarationUsage.BiNormal, 0),
			    VertexElement.VertexDeclarationEnd,
		    };
            #endregion

            #region bV̍쐬
            Mesh mesh = new Mesh(
					12, // Op`̐
					36, // _̐
					MeshFlags.Managed,
					elements,
					this.device );

            //_obt@ɏ݊Jn
			VertexBuffer vb = mesh.VertexBuffer;
            GraphicsStream vbuf = vb.Lock(0,0,LockFlags.None);
            vbuf.Write(verts);
            vb.Unlock();

            //obt@ɏ݊Jn
			IndexBuffer ib = mesh.IndexBuffer;
            GraphicsStream ibuf = ib.Lock(0,0,LockFlags.None);
            for(int n=0;n<verts.Length;n++)
            {
                ibuf.Write((Int16)n);
            }
			ib.Unlock();
            #endregion

            return mesh;
        }

        /// <summary>
        /// ɕ`R[h
        /// </summary>
        public void Render(bool bUseNormalMap)
        {
            if (this.device == null) return;

            // `eŃNAAyobt@NAB
            this.device.Clear(ClearFlags.ZBuffer | ClearFlags.Target,
                Color.Black, 1.0f, 0);


            // At@uhL
            this.device.RenderState.AlphaBlendEnable = true;
            this.device.RenderState.SourceBlend = Blend.SourceAlpha;
            this.device.RenderState.DestinationBlend = Blend.InvSourceAlpha;

            // GtFNgĂȂȂ炱̕`͒߂
            if(effect==null) return;

            // uBeginScenevƁuEndScenev̊Ԃɕ`eLq
            this.device.BeginScene();


            // eNjbNw
            effect.Technique = effect.GetTechnique("FixedColor");

            // r{[h̕\
            device.RenderState.ShadeMode = ShadeMode.Gouraud;
            device.RenderState.CullMode = Cull.None;
            device.VertexFormat = CustomVertex.TransformedColored.Format;
            device.DrawIndexedUserPrimitives(PrimitiveType.TriangleList,
                0, 6, 2, backgroundIndices, true, backgroundVertices);

    		long	m_psec,m_time;
            QueryPerformanceFrequency(out m_psec);
            QueryPerformanceCounter(out m_time);
            double passedSecond = (double)(m_time) / (double)m_psec;

            ///// [JWniIuWFNgj̐ݒ /////
            Matrix mW =
                Matrix.RotationY((float)passedSecond * 0.2f)
              * Matrix.RotationX(-(float)Math.PI * 0.15f);

            ///// O[oWni_j̐ݒ /////
            Matrix mV =
                Matrix.Translation(0,-1,20);
            Matrix mP =
                Matrix.PerspectiveFovLH(
                    Geometry.DegreeToRadian(30.0f),
                    (float)presentParameters.BackBufferWidth / (float)presentParameters.BackBufferHeight,
                    0.01f, 100.0f);

            // eNjbNw
            effect.Technique = effect.GetTechnique("GradientNormalMapping");

            // }bsOtO
            effect.SetValue("pMapping",new Vector4(g_baseTexture!=null ? 1:0,g_nmhTexture!=null ? 1:0,0,0));

            // eNX`Zbg
            effect.SetValue("g_baseTexture", g_baseTexture);

            effect.SetValue("g_nmhTexture", g_nmhTexture);

	        // O[oԂł̃Cgʒu
	        effect.SetValue("vLightWS", new Vector4(0,0,-1024,0) );
            
            // Of[V
            effect.SetValue("vGradient", new Vector4(0, 1, 0, 0));

            // XP[Zbgi摜̃AXyNglj
            float wph = (float)aspect.Width / (float)aspect.Height;
            float xScale = (wph>=1) ? 2 : 2*wph;
            float yScale = (wph>=1) ? 2/wph : 2;
            effect.SetValue("vScale", new Vector4(xScale,yScale,xScale, 1));

            ////////////////////////////////////////////////// 
            // ˂bVɕ\

            // F̐ݒ
            effect.SetValue("cBase0", new Vector4(1, 1, 1, 0));
            effect.SetValue("cBase1", new Vector4(1, 1, 1, 0.5f));

            // Of[VW
            effect.SetValue("pGradient", new Vector4(-0.5f, 1, 0, 0));

            // ]WZbgix]j
            effect.SetValue("vInverse", new Vector4(1, -1, 1, 1));

            // ː̐ݒ
            float fSpecular = bUseNormalMap ? 0.5f : 0.0f;
            effect.SetValue("pReflect", new Vector4(0.2f, 0.8f, fSpecular, 5.0f));

            // JO𔽎vɐݒ
            device.RenderState.CullMode = Cull.Clockwise;

            // ˂ꂽ{bNX̕`
            drawMesh(boxMesh, Matrix.Translation(0, -yScale, 0) * mW, mV, mP);

            ////////////////////////////////////////////////// 
            // bVɕ\

            // F̐ݒ
            effect.SetValue("cBase0", new Vector4(1, 1, 1, 1));
            effect.SetValue("cBase1", new Vector4(1, 1, 1, 1));

            // Of[VW
            effect.SetValue("pGradient", new Vector4(-1, 1, 0, 0));

            // ]WZbgi]Ȃj
            effect.SetValue("vInverse", new Vector4(1, 1, 1, 1));

            // JO𔽎vɐݒ
            device.RenderState.CullMode = Cull.CounterClockwise;

            // {bNX̕`
            drawMesh(boxMesh, Matrix.Translation(0, yScale, 0) * mW, mV, mP);

            // ̕\
            if(!string.IsNullOrEmpty(usage))
            {
                font.DrawText(null, usage, 8, presentParameters.BackBufferHeight-20, Color.White);
            }

            this.device.EndScene();


            #region V[̍XV
            try
            {
                this.device.Present();
            }
            // foCX̃`FbNƃZbg
            catch (DeviceLostException)
            {
                int result;
                // Zbg\ǂ`FbN
                if (!this.device.CheckCooperativeLevel(out result))
                {
                    // Zbg\Ȃ烊Zbg
                    if (result == (int)ResultCode.DeviceNotReset)
                        this.device.Reset(presentParameters);
                    // ܂ZbgłȂ΁A΂炭X[v
                    else if (result == (int)ResultCode.DeviceLost)
                        System.Threading.Thread.Sleep(20);
                }
            }
            #endregion
        }

        /// <summary>
        /// bV̕`
        /// </summary>
        /// <param name="mesh">bV</param>
        /// <param name="wM">[hWn</param>
        /// <param name="vM">r[Wn</param>
        /// <param name="pM">eWn</param>
        /// <param name="scale">XP[</param>
        protected void drawMesh(Mesh mesh,Matrix wM,Matrix vM,Matrix pM)
        {
            // Wϊs̐ݒ
            effect.SetValue("mW", wM);
            effect.SetValue("mWV", wM * vM);
            effect.SetValue("mWVP", wM * vM * pM);

            int nPasses = 1;
            nPasses = effect.Begin(0);

            // _Oi`j
            for (int i = 0; i < nPasses; i++)
            {
                effect.BeginPass(i);
                effect.CommitChanges();
                mesh.DrawSubset( 0 );
                effect.EndPass();
            }

            effect.End();
        }

        #region Win32 API
        [System.Security.SuppressUnmanagedCodeSecurity]
        [DllImport("kernel32")]
        public static extern bool QueryPerformanceFrequency(out long PerformanceFrequency);

        [System.Security.SuppressUnmanagedCodeSecurity]
        [DllImport("kernel32")]
        public static extern bool QueryPerformanceCounter(out long PerformanceCount);
        #endregion
    }
}
