/////////////////////////////////////////////////////////
//
// Gradient Normal Mapping Shader
//
// Copyright 2007 Shintaro Takemura
//

// O[oϐ //

// J[eNX`
texture g_baseTexture;

// @}bv
texture g_nmhTexture;

// J[Tv[
sampler2D tBase = sampler_state
{
	Texture = < g_baseTexture >;
	MinFilter = ANISOTROPIC;
	MagFilter = Linear;
	MipFilter = Linear;
	AddressU = Wrap;
	AddressV = Wrap;
	MaxAnisotropy = 16;
};

// @Tv[
sampler2D tNormalHeightMap = sampler_state
{
	Texture = < g_nmhTexture >;
	MinFilter = ANISOTROPIC;
	MagFilter = Linear;
	MipFilter = Linear;
	AddressU = Wrap;
	AddressV = Wrap;
	MaxAnisotropy = 16;
};

// x[XJ[O
float4 cBase0;

// x[XJ[P
float4 cBase1;

// }bsO̗LtO(R=eNX`,G=@}bvj
float4 pMapping;

// ː(R=gU,G=͌,B=ˌ,A=˂̋)
float4 pReflect;

// Of[VJ[uiR=,G=j
float4 pGradient;

// Of[V
float4 vGradient;

// bṼXP[
float4 vScale;

// bV̔]
float4 vInverse;

// Cg̈ʒu
float4 vLightWS;

// Wϊs World
float4x4 mW;

// Wϊs World * View
float4x4 mWV;

// Wϊs World * View * Projection
float4x4 mWVP;

// sNZVF[_[͍̓\
struct PS_INPUT
{
	float4 position          : POSITION;
	float2 texCoord          : TEXCOORD0;
	float3 vLightTS          : TEXCOORD1;
	float3 vViewTS           : TEXCOORD2;
	float3 vNormalWS         : TEXCOORD4;
	float3 vViewWS           : TEXCOORD5;
	float3 vPositionOS       : TEXCOORD6;
};

////////////////////////////////////////
// _VF[_[ //
PS_INPUT VS_TangentVectorSpace (
	float4 inPositionOS : POSITION,
	float3 vInTangentOS : TANGENT,
	float3 vInNormalOS : NORMAL,
	float3 vInBinormalOS : BINORMAL
	)
{
	// sNZVF[_[͍̓\̂錾
	PS_INPUT Out;

	// _W̃XP[O
	float4 vScaledPos = inPositionOS * vScale * vInverse;
	Out.vPositionOS = inPositionOS * vInverse;

	// Wϊ̒_ʒuo
	Out.position = mul(vScaledPos, mWVP);

	// ڐxNgƈʒu̓ςtWo
	// ]@xNgƈʒu̓ςuWo
	Out.texCoord = float2(
	 dot(vInTangentOS , inPositionOS) * 0.5 + 0.5,
	 dot(vInBinormalOS, inPositionOS) * -0.5 + 0.5 );

	// O[oԂł̈ʒuZo
	float4 vPositionWS = mul( inPositionOS, mW );

	// O[oԂł̐ڋԊZo
	float3 vNormalWS   = mul( vInNormalOS  , (float3x3) mW );
	float3 vTangentWS  = mul( vInTangentOS , (float3x3) mW );
	float3 vBinormalWS = mul( vInBinormalOS, (float3x3) mW );

	// O[oԂł̎xNgo
	float3 vViewWS = -mul( inPositionOS, mWV );
	Out.vViewWS = vViewWS;

	// O[oԂł̖@xNgo
	Out.vNormalWS = vNormalWS;

	// ڐE@E]@𐳋K
	vNormalWS   = normalize( vNormalWS );
	vTangentWS  = normalize( vTangentWS );
	vBinormalWS = normalize( vBinormalWS );

	// ڋԍWϊsZo
	float3x3 mWorldToTangent = float3x3( vTangentWS, vBinormalWS, vNormalWS );

	// ڃxNgԂł̃CgxNgo
	Out.vLightTS = mul( vLightWS, mWorldToTangent );

	// ڃxNgԂł̎xNgo
	Out.vViewTS  = mul( mWorldToTangent, vViewWS  );

	return Out;
}

// sNZVF[_[ //
float4 PS_GradientNormalMapping( PS_INPUT i ) : COLOR0
{
	//  ԂꂽxNg𐳋K
	float3 vViewTS   = normalize( i.vViewTS  );
	float3 vLightTS  = normalize( i.vLightTS );

	// ڃxNgԂɂ@xNg`
	float3 vNormalTS = pMapping.y>=1 ? normalize( tex2D( tNormalHeightMap, i.texCoord ) * 2 - 1 ) : float3(0,0,1);
	
	// Of[VEFCgZo
	float t = dot(i.vPositionOS,vGradient);
	float a = smoothstep(pGradient.x,pGradient.y,t);
	float4 cGradColor = lerp(cBase0,cBase1,a);

	// J[eNX`eNZ擾
	float4 cTexelColor = pMapping.x>=1 ? tex2D( tBase, i.texCoord ) : float4(1,1,1,1);

	// gUːZo
	float tDiffuse = saturate( dot( vNormalTS, vLightTS ));
	float d = tDiffuse * pReflect.g + pReflect.r ;

	// ʔːZo
	float3 vReflectionTS = normalize( 2 * dot( vViewTS, vNormalTS ) * vNormalTS - vViewTS);
	float fRdotL = saturate( dot( vReflectionTS, vLightTS ));
	float s = saturate( pow( fRdotL, pReflect.a )) * pReflect.b;

	// ŏIo͌ʂZo
	return ( float4(d,d,d,1) * cGradColor * cTexelColor + float4(s,s,s,0) );
}

////////////////////////////////////////
// Of[V{@}bsO
technique GradientNormalMapping
{
	pass P0
	{
		VertexShader = compile vs_2_0 VS_TangentVectorSpace();
		PixelShader  = compile ps_2_0 PS_GradientNormalMapping();
	}
}

////////////////////////////////////////
// ŒFhԂ

technique FixedColor
{
    pass P0
    {          
		VertexShader = null;
		PixelShader  = null;
    }
}